@@ -21,16 +21,8 @@
  */
 
 
-#ifdef HAVE_CONFIG_H
-#  include <config.h>
-#endif
-
-#include <gio/gio.h>
-#include <glib/gi18n-lib.h>
 #include "player.h"
-#include "stream.h"
 
-static gint64   dur;
 static gboolean loop_stream_was_played = FALSE; ///< endless loop prevention flag
 
 /**
@@ -42,9 +34,10 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
   {
     case GST_MESSAGE_EOS:
     {
+      stream_set_paused (*currentstream, FALSE);
       stream_print_info (*currentstream);
       g_debug ("%s: End-of-Stream", __FUNCTION__);
-      player_prepare_for_next ();
+      play_next ();
       break;
     }
     
@@ -57,7 +50,7 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
       g_warning ("Player: %s. Skipping...", err->message);
       g_error_free (err);
       
-      player_prepare_for_next ();
+      play_next ();
       
       break;
     }
@@ -94,25 +87,17 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
           gst_element_state_get_name (newstate),
           gst_element_state_get_name (targetstate));
       
-      if (newstate == GST_STATE_READY)
-      {
-        /* This happens when:
-         *  - when we need to play the next stream in queue.
-         *    (NULL -> READY)
-         * 
-         * We ask to play the next stream in queue.
-         */
-        g_debug ("%s: playing next because GST_STATE_READY", __FUNCTION__);
-        play_next ();
-      }
-      else if (oldstate == GST_STATE_READY && newstate == GST_STATE_PAUSED)
+      if (oldstate == GST_STATE_READY && newstate == GST_STATE_PAUSED)
       {
         /* This happens when:
          *  - stream is getting ready to play (READY->PAUSED)
          * 
          * We need to seek/forward to the wanted position (ap_config.position).
          */
-        player_seek ();
+        player_set_position (ap_config.position);
+        stream_set_paused (*currentstream, TRUE);
+        // Request to start playing.
+        player_start_playing();
         stream_add_updater (*currentstream, 
                             (GSourceFunc) player_update_stream);
       }
@@ -120,7 +105,7 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
       {
         /* This happens when:
          *  - stream has just started playing (PAUSED->PLAYING)
-         *  - stream has resumed from user-requested-pause (PAUSED->PLAYING)
+         *  - stream has resumed from user-requested pause (PAUSED->PLAYING)
          *  - (maybe) when buffer was empty.
          */
           
@@ -128,13 +113,14 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
         
         if (!loop_stream_was_played)
           loop_stream_was_played = TRUE;
-        
+        stream_set_paused (*currentstream, FALSE);
         stream_print_info (*currentstream);
       }
       else if (oldstate == GST_STATE_PLAYING && newstate == GST_STATE_PAUSED)
       {
         g_debug ("%s: Paused!", __FUNCTION__);
         
+        stream_set_paused (*currentstream, TRUE);
         stream_print_info (*currentstream);
       }
       
@@ -196,60 +182,44 @@ gboolean bus_call (GstBus *bus, GstMessage *msg, gpointer user_data)
  */
 void play_next ()
 {
-  gboolean need_newline = FALSE;
+  if (!player_get_ready())
+    return;
+  
+  gboolean need_newline = TRUE;
   g_debug ("%s: Playing next...", __FUNCTION__);
   
-  // if just started playing
-  if (!ap_config.current)
+  // get the next stream from the list
+  Stream* next_stream = playlist_next();
+  
+  if (!current_stream)
+    need_newline = FALSE;
+  
+  // if Loop && we are at the end && a stream was just played, we go back to the first one
+  if (ap_config.loop && !next_stream)
   {
-    ap_config.glist_current = ap_config.playlist;
-  }
-  else // else, we get the next stream
-  {
-    // we need a new line between streams
-    need_newline = TRUE;
-    
-    // get the next stream from the list
-    ap_config.glist_current = g_list_next (ap_config.glist_current);
-    
-    // if Loop && we are at the end && a stream was just played, we go back to the first one
-    if (ap_config.loop && !ap_config.glist_current)
+    // check for fast endless loop
+    if (loop_stream_was_played)
     {
-      if (loop_stream_was_played)
-      {
-        g_debug ("Looping...");
-        loop_stream_was_played = FALSE;
-        ap_config.glist_current = ap_config.playlist;
-      }
-      else
-      {
-        g_warning ("Deactivating loop mode: useless endless loop detected.");
-        ap_config.loop = FALSE;
-      }
+      g_debug ("Looping...");
+      loop_stream_was_played = FALSE;
+      next_stream = playlist_first ();
+    }
+    else
+    {
+      g_warning ("Deactivating loop mode: fast endless loop detected.");
+      ap_config.loop = FALSE;
     }
   }
   
-  // if we have a current stream
-  if (ap_config.glist_current)
+  // if we have a stream
+  if (next_stream)
   {
-    ap_config.current = ap_config.glist_current->data;
-    
     // we print the new line now so that it's in between streams (to seperate them)
     if (!ap_config.debug && need_newline)
       g_print("\n");
-    
-    // Define the URI to play
-    g_debug ("%s: URI: %s", __FUNCTION__, (*currentstream)->uri);
-    g_object_set (G_OBJECT (pipeline), "uri", (*currentstream)->uri, NULL);
-    
-    // Show "Loading Filename..."
-    g_info (_("Loading \"%s\"...\n"),
-            (*currentstream)->filepath ?
-              (*currentstream)->filepath : (*currentstream)->uri
-           );
   
     // Starts playing
-    player_begin ();
+    player_begin (next_stream);
   }
   else // if no current stream, we clear the pipeline and call the callback
   {
@@ -258,6 +228,45 @@ void play_next ()
   }
 }
 
+/**
+ * Play the next stream in the playlist. (asynchronous)
+ */
+void play_previous ()
+{
+  if (!player_get_ready())
+    return;
+  
+  gboolean need_newline = TRUE;
+  g_debug ("%s: Playing next...", __FUNCTION__);
+  
+  // get the next stream from the list
+  Stream* prev_stream = playlist_previous ();
+  
+  // if Loop && we are at the end && a stream was just played, we go back to the first one
+  if (ap_config.loop && !prev_stream)
+  {
+    g_debug ("Looping...");
+    loop_stream_was_played = FALSE;
+    prev_stream = playlist_last ();
+  }
+  
+  // if we have a stream
+  if (prev_stream)
+  {
+    // we print the new line now so that it's in between streams (to seperate them)
+    g_print("\n");
+  
+    // Starts playing
+    player_begin (prev_stream);
+  }
+  else
+  {
+    // if we have no stream, we play the first
+    player_begin (playlist_first ());
+  }
+  
+}
+
 /**
  * Initialize the player.
  * This initializes the player before starting any play. This should be called
@@ -308,16 +317,15 @@ void player_init (void *callback_when_done, void *bus_call_dispatcher)
   done_playing_callback = callback_when_done;
   
   // currentstream is a pointer to pointer to stream
-  currentstream = &ap_config.current;
-  player_prepare_for_next ();
+  currentstream = &current_stream;
+  play_next ();
 }
 
 /**
- * Asks the player to prepare itself to play the next stream in queue.
- * This asks the player to prepare itself to play the next stream in queue. The
- * next stream will be played when the pipeline is in READY state.
+ * Asks the player to prepare itself to play another stream.
+ * This asks the player to prepare itself to play another stream.
  */
-void player_prepare_for_next ()
+gboolean player_get_ready ()
 {
   if (currentstream && *currentstream)
     stream_remove_updater (*currentstream); 
@@ -328,15 +336,15 @@ void player_prepare_for_next ()
     gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_NULL);
   
   gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_READY);
-}
-
-/**
- * Same as player_prepare_for_next () except this returns FALSE (for g_idle).
- */
-static gboolean player_prepare_for_next_query ()
-{
-  player_prepare_for_next ();
-  return FALSE;
+  if (gst_element_get_state (GST_ELEMENT (pipeline), NULL, NULL, 5 * GST_SECOND) == GST_STATE_CHANGE_SUCCESS)
+  {
+    return TRUE;
+  }
+  else
+  {
+    g_error (_("Couldn't get the player ready"));
+    return FALSE;
+  }
 }
 
 /**
@@ -360,10 +368,24 @@ void player_clear_pipeline ()
 }
 
 /**
- * Asks the player to be in PAUSED mode. (might be asynchronous)
+ * Asks the player to play the requested stream
+ * Actually this asks the player to be in PAUSED mode
+ * (to be able to seek afterwards).
  */
-static void player_begin ()
+static void player_begin (Stream* stream)
 {
+  if (!stream)
+    return;
+  
+  current_stream = stream;
+  
+  // Define the URI to play
+  g_debug ("%s: URI: %s", __FUNCTION__, stream->uri);
+  g_object_set (G_OBJECT (pipeline), "uri", stream->uri, NULL);
+  
+  // Show "Loading Filename..."
+  g_info (_("Loading \"%s\"...\n"), stream->uri);
+  
   g_debug ("%s: PAUSED state: requesting...", __FUNCTION__);
   switch(gst_element_set_state (GST_ELEMENT (pipeline), GST_STATE_PAUSED))
   {
@@ -374,7 +396,7 @@ static void player_begin ()
                   (*currentstream)->filepath ?
                    (*currentstream)->filepath : (*currentstream)->uri
                 );
-      player_prepare_for_next ();
+      play_next ();
       return;
       break;
     }
@@ -406,7 +428,7 @@ static void player_begin ()
 /**
  * Seek to the necessary position (depending on ap_config). (synchronous)
  */
-static void player_seek ()
+void player_set_position (StartPlayingPosition position)
 {
   g_debug ("%s: Checking duration", __FUNCTION__);
   
@@ -414,11 +436,9 @@ static void player_seek ()
   if (player_update_stream_duration (*currentstream))
   {
     // -If- negative or null duration (possibly a continuous stream)
-    // -OR- Play from the start.
-    if (ap_config.position == START_PLAYING_START || (*currentstream)->duration <= 0)
+    if ((*currentstream)->duration <= 0)
     {
-      if ((*currentstream)->duration <= 0)
-        g_debug ("%s: No seek: duration is 0...", __FUNCTION__);
+      g_debug ("%s: No seek: duration is 0...", __FUNCTION__);
     }
     else // -else- Seek to requested position (duration is in nanosec)
     {
@@ -437,12 +457,12 @@ static void player_seek ()
       gint64 seekposition = 0;
       
       // -If- play from MIDDLE
-      if (ap_config.position == START_PLAYING_MIDDLE)
+      if (position == START_PLAYING_MIDDLE)
       {
         // seek position = the middle of the duration
         seekposition = (*currentstream)->duration/2;
       }
-      else if (ap_config.position == START_PLAYING_END) // -If- Play the end
+      else if (position == START_PLAYING_END) // -If- Play the end
       {
         gint64 configduration = ap_config.duration*GST_SECOND;
         // -If- wanted duration is shorter than stream duration, we can seek.
@@ -456,7 +476,7 @@ static void player_seek ()
                    __FUNCTION__);
         }
       }
-      else if (ap_config.position == START_PLAYING_ANYWHERE) 
+      else if (position == START_PLAYING_ANYWHERE) 
       { // -If- play from ANYWHERE
       
         gint streamsec = (*currentstream)->duration/GST_SECOND; // in sec
@@ -479,33 +499,36 @@ static void player_seek ()
           g_debug ("%s: wanted duration is longer than stream", __FUNCTION__);
       }
       
-      // -If- we need to seek (seekpositon != 0)
-      if (seekposition)
-      {
-        g_debug ("%s: Seeking...", __FUNCTION__);
-        
-        GstSeekFlags seekflags = GST_SEEK_FLAG_FLUSH;
-        if (ap_config.fastseek)
-        {
-          g_debug ("%s: Fast seek.", __FUNCTION__);
-          seekflags |= GST_SEEK_FLAG_KEY_UNIT;
-        }
-        else
-          seekflags |= GST_SEEK_FLAG_ACCURATE;
-        
-        if (!gst_element_seek_simple (pipeline, GST_FORMAT_TIME, seekflags,
-                                      seekposition))
-          g_warning ("%s: Unable to seek stream", __FUNCTION__);
-      }
+      /* we seek to the desired position */
+      player_seek (seekposition);
+      
     }
   }
   else
   {
     g_warning ("%s: Couldn't find stream duration...", __FUNCTION__);
   }
+}
+
+void player_seek (gint64 seekposition)
+{
+  g_debug ("%s: Seeking...", __FUNCTION__);
   
-  // Request to start playing.
-  player_start_playing();
+  GstSeekFlags seekflags = GST_SEEK_FLAG_FLUSH;
+  if (ap_config.fastseek)
+  {
+    g_debug ("%s: Fast seek.", __FUNCTION__);
+    seekflags |= GST_SEEK_FLAG_KEY_UNIT;
+  }
+  else
+    seekflags |= GST_SEEK_FLAG_ACCURATE;
+  
+  if (!gst_element_seek_simple (pipeline, GST_FORMAT_TIME, seekflags,
+                                seekposition))
+    g_warning ("%s: Unable to seek stream", __FUNCTION__);
+  
+  // we reset the start position of the stream
+  (*currentstream)->startposition = -1;
 }
 
 /**
@@ -556,18 +579,18 @@ gboolean player_update_stream ()
   
   stream_print_info (*currentstream);
   
-  if ((*currentstream)->startposition == -1)
+  if ((*currentstream)->startposition == -1 )
     (*currentstream)->startposition = (*currentstream)->position;
+    
+  // UNLOCK
+  g_static_mutex_unlock (&mutex);
   
   stream_get_elapsed_time(*currentstream, &elapsed);
   
   /* If we have reached Duration or Maxduration, we play the next stream */
   if (ap_config.duration >= 0 && elapsed >= (gint64)ap_config.duration*GST_SECOND
    || ap_config.maxduration >= 0 && elapsed >= (gint64)ap_config.maxduration*GST_SECOND)
-    player_prepare_for_next();
-  
-  // UNLOCK
-  g_static_mutex_unlock (&mutex);
+    play_next ();
   
   return TRUE;
 }
