@@ -24,8 +24,15 @@
 #define PLAYLIST_H
 
 #include <gst/gst.h>
+#include "ap_config.h"
+#include "functions.h"
+#include "stream.h"
 
 gboolean        playlist_add              (const gchar* path);
+Stream*         playlist_next             ();
+Stream*         playlist_previous         ();
+Stream*         playlist_first            ();
+Stream*         playlist_last             ();
 void            playlist_shuffle          ();
 void            playlist_free             ();
 gboolean        playlist_add_from_file    (const gchar* path);
